@@ -698,6 +698,13 @@ static inline void insert_driver(struct pw_context *context, struct pw_impl_node
 			break;
 	}
 	spa_list_append(&n->driver_link, &node->driver_link);
+	pw_context_emit_driver_added(context, node);
+}
+
+static inline void remove_driver(struct pw_context *context, struct pw_impl_node *node)
+{
+	spa_list_remove(&node->driver_link);
+	pw_context_emit_driver_removed(context, node);
 }
 
 static void update_io(struct pw_impl_node *node)
@@ -740,6 +747,7 @@ int pw_impl_node_register(struct pw_impl_node *this,
 		PW_KEY_MODULE_ID,
 		PW_KEY_FACTORY_ID,
 		PW_KEY_CLIENT_ID,
+		PW_KEY_CLIENT_API,
 		PW_KEY_DEVICE_ID,
 		PW_KEY_PRIORITY_SESSION,
 		PW_KEY_PRIORITY_DRIVER,
@@ -766,6 +774,7 @@ int pw_impl_node_register(struct pw_impl_node *this,
 	this->global = pw_global_new(context,
 				     PW_TYPE_INTERFACE_Node,
 				     PW_VERSION_NODE,
+				     PW_NODE_PERM_MASK,
 				     properties,
 				     global_bind,
 				     this);
@@ -846,8 +855,8 @@ do_move_nodes(struct spa_loop *loop,
 static void remove_segment_owner(struct pw_impl_node *driver, uint32_t node_id)
 {
 	struct pw_node_activation *a = driver->rt.target.activation;
-	ATOMIC_CAS(a->segment_owner[0], node_id, 0);
-	ATOMIC_CAS(a->segment_owner[1], node_id, 0);
+	SPA_ATOMIC_CAS(a->segment_owner[0], node_id, 0);
+	SPA_ATOMIC_CAS(a->segment_owner[1], node_id, 0);
 }
 
 SPA_EXPORT
@@ -941,8 +950,9 @@ static void check_properties(struct pw_impl_node *node)
 		if (node->registered) {
 			if (driver)
 				insert_driver(context, node);
-			else
-				spa_list_remove(&node->driver_link);
+			else {
+				remove_driver(context, node);
+			}
 		}
 		if (driver && node->driver_node == node)
 			node->driving = true;
@@ -1158,7 +1168,7 @@ static inline int trigger_targets(struct pw_impl_node *this, int status, uint64_
 		pw_log_trace_fp("%p: (%s-%u) state:%p pending:%d/%d", t->node,
 				t->name, t->id, state, state->pending, state->required);
 
-		if (pw_node_activation_state_dec(state, 1)) {
+		if (pw_node_activation_state_dec(state)) {
 			a->status = PW_NODE_ACTIVATION_TRIGGERED;
 			a->signal_time = nsec;
 			if (SPA_UNLIKELY(spa_system_eventfd_write(t->system, t->fd, 1) < 0))
@@ -1249,11 +1259,12 @@ static inline int process_node(void *data)
 		/* calculate CPU time when finished */
 		a->signal_time = this->driver_start;
 		calculate_stats(this, a);
-		pw_context_driver_emit_complete(this->context, this);
+		pw_impl_node_rt_emit_complete(this);
+//		pw_context_driver_emit_complete(this->context, this);
 	}
 
 	if (SPA_UNLIKELY(status & SPA_STATUS_DRAINED))
-		pw_context_driver_emit_drained(this->context, this);
+		pw_impl_node_rt_emit_drained(this);
 
 	return status;
 }
@@ -1263,7 +1274,7 @@ int pw_impl_node_trigger(struct pw_impl_node *node)
 	struct pw_node_activation *a = node->rt.target.activation;
 	struct pw_node_activation_state *state = &a->state[0];
 
-	if (pw_node_activation_state_dec(state, 1)) {
+	if (pw_node_activation_state_dec(state)) {
 		uint64_t nsec = get_time_ns(node->data_system);
 		a->status = PW_NODE_ACTIVATION_TRIGGERED;
 		a->signal_time = nsec;
@@ -1394,6 +1405,7 @@ struct pw_impl_node *pw_context_create_node(struct pw_context *context,
 	spa_list_init(&this->peer_list);
 
 	spa_hook_list_init(&this->listener_list);
+	spa_hook_list_init(&this->rt_listener_list);
 
 	this->info.state = PW_NODE_STATE_CREATING;
 	this->info.props = &this->properties->dict;
@@ -1661,8 +1673,8 @@ static inline int check_updates(struct pw_impl_node *node, uint32_t *reposition_
 	if (SPA_UNLIKELY(a->position.offset == INT64_MIN))
 		a->position.offset = a->position.clock.position;
 
-	command = ATOMIC_XCHG(a->command, PW_NODE_ACTIVATION_COMMAND_NONE);
-	*reposition_owner = ATOMIC_XCHG(a->reposition_owner, 0);
+	command = SPA_ATOMIC_XCHG(a->command, PW_NODE_ACTIVATION_COMMAND_NONE);
+	*reposition_owner = SPA_ATOMIC_XCHG(a->reposition_owner, 0);
 
 	if (SPA_UNLIKELY(command != PW_NODE_ACTIVATION_COMMAND_NONE)) {
 		pw_log_debug("%p: update command:%u", node, command);
@@ -1726,7 +1738,7 @@ static inline void update_position(struct pw_impl_node *node, int all_ready, uin
 			pw_log_warn("(%s-%u) sync timeout, going to RUNNING",
 					node->name, node->info.id);
 			check_states(node, nsec);
-			pw_context_driver_emit_timeout(node->context, node);
+			pw_impl_node_rt_emit_timeout(node);
 			all_ready = true;
 		}
 		if (all_ready)
@@ -1775,7 +1787,7 @@ static int node_ready(void *data, int status)
 					state, a->position.clock.duration,
 					state->pending, state->required);
 			check_states(node, nsec);
-			pw_context_driver_emit_incomplete(node->context, node);
+			pw_impl_node_rt_emit_incomplete(node);
 		}
 
 		/* This update is done too late, the driver should do this
@@ -1790,8 +1802,8 @@ static int node_ready(void *data, int status)
 		}
 
 		sync_type = check_updates(node, &reposition_owner);
-		owner[0] = ATOMIC_LOAD(a->segment_owner[0]);
-		owner[1] = ATOMIC_LOAD(a->segment_owner[1]);
+		owner[0] = SPA_ATOMIC_LOAD(a->segment_owner[0]);
+		owner[1] = SPA_ATOMIC_LOAD(a->segment_owner[1]);
 again:
 		all_ready = sync_type == SYNC_CHECK;
 		update_sync = !all_ready;
@@ -1841,7 +1853,7 @@ again:
 
 		update_position(node, all_ready, nsec);
 
-		pw_context_driver_emit_start(node->context, node);
+		pw_impl_node_rt_emit_start(node);
 	}
 	/* this should not happen, driver nodes that are not currently driving
 	 * should not emit the ready callback */
@@ -1904,7 +1916,7 @@ static int node_xrun(void *data, uint64_t trigger, uint64_t delay, struct spa_po
 				missed);
 	}
 
-	pw_context_driver_emit_xrun(this->context, this);
+	pw_impl_node_rt_emit_xrun(this);
 
 	return 0;
 }
@@ -1954,6 +1966,50 @@ void pw_impl_node_add_listener(struct pw_impl_node *node,
 	spa_hook_list_append(&node->listener_list, listener, events, data);
 }
 
+struct listener_data {
+	struct spa_hook *listener;
+	const struct pw_impl_node_rt_events *events;
+	void *data;
+};
+
+static int
+do_add_rt_listener(struct spa_loop *loop,
+		bool async, uint32_t seq, const void *data, size_t size, void *user_data)
+{
+	struct pw_impl_node *node = user_data;
+	const struct listener_data *d = data;
+	spa_hook_list_append(&node->rt_listener_list,
+			d->listener, d->events, d->data);
+	return 0;
+}
+
+SPA_EXPORT
+void pw_impl_node_add_rt_listener(struct pw_impl_node *node,
+			   struct spa_hook *listener,
+			   const struct pw_impl_node_rt_events *events,
+			   void *data)
+{
+	struct listener_data d = { .listener = listener, .events = events, .data = data };
+	pw_loop_invoke(node->data_loop,
+                       do_add_rt_listener, SPA_ID_INVALID, &d, sizeof(d), false, node);
+}
+
+static int do_remove_listener(struct spa_loop *loop,
+		bool async, uint32_t seq, const void *data, size_t size, void *user_data)
+{
+	struct spa_hook *listener = user_data;
+	spa_hook_remove(listener);
+	return 0;
+}
+
+SPA_EXPORT
+void pw_impl_node_remove_rt_listener(struct pw_impl_node *node,
+			  struct spa_hook *listener)
+{
+	pw_loop_invoke(node->data_loop,
+                       do_remove_listener, SPA_ID_INVALID, NULL, 0, true, listener);
+}
+
 /** Destroy a node
  * \param node a node to destroy
  *
@@ -1998,7 +2054,7 @@ void pw_impl_node_destroy(struct pw_impl_node *node)
 	if (node->registered) {
 		spa_list_remove(&node->link);
 		if (node->driver)
-			spa_list_remove(&node->driver_link);
+			remove_driver(context, node);
 	}
 
 	if (node->node) {
